import os
from collections import OrderedDict
from math import ceil, floor

import cdsapi
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import ticker
from matplotlib.colors import Normalize
from netCDF4 import Dataset
from numba import float64
from numba.experimental import jitclass
from numba.types import unicode_type

from ..queryreport import print_text
from ..utilities.utilities import SA_Density, SA_Pressure, SA_Temperature, interp_ex

MetData_spec = OrderedDict()
MetData_spec["z_data"] = float64[::1]
MetData_spec["temperature_data"] = float64[::1]
MetData_spec["pressure_data"] = float64[::1]
MetData_spec["density_data"] = float64[::1]
MetData_spec["wind_U_data"] = float64[::1]
MetData_spec["wind_V_data"] = float64[::1]
MetData_spec["source"] = unicode_type
MetData_spec["surface_temperature"] = float64  # surface temperature
MetData_spec["surface_pressure"] = float64  # surface pressure
MetData_spec["lapse_tropos"] = float64  # lapse rate in troposphere
MetData_spec["lapse_stratos"] = float64  # lapse rate stratosphere
MetData_spec["height_tropos"] = float64  # height of the troposphere
MetData_spec["height_stratos"] = float64  # height of the stratosphere
MetData_spec["Ra"] = float64  # gas constant of dry air
MetData_spec["g"] = float64  # gravitational acceleration


@jitclass(MetData_spec)
class MetData:
    def __init__(
        self,
        source="standardAtmos",
        surface_temperature=293,
        surface_pressure=101325,
        lapse_tropos=6.5e-3,
        lapse_stratos=2.0e-3,
        height_tropos=11e3,
        height_stratos=20e3,
        Ra=285,
        g=9.81,
    ):
        self.surface_temperature = np.float64(surface_temperature)
        self.surface_pressure = np.float64(surface_pressure)
        self.lapse_tropos = np.float64(lapse_tropos)
        self.lapse_stratos = np.float64(lapse_stratos)
        self.height_tropos = np.float64(height_tropos)
        self.height_stratos = np.float64(height_stratos)
        self.Ra = np.float64(Ra)
        self.g = np.float64(g)
        self.source = source

        self.z_data = np.empty((0), dtype=np.float64)
        self.temperature_data = np.empty((0), dtype=np.float64)
        self.pressure_data = np.empty((0), dtype=np.float64)
        self.density_data = np.empty((0), dtype=np.float64)
        self.wind_U_data = np.empty((0), dtype=np.float64)
        self.wind_V_data = np.empty((0), dtype=np.float64)

    def set_zuv_data(self, z, u, v):
        self.z_data = z
        self.wind_U_data = u
        self.wind_V_data = v

    def temperature(self, z):
        if self.source in ["netCDF", "GFS"]:
            temperature = interp_ex(z, self.z_data, self.temperature_data)
        if self.source == "standardAtmos":
            temperature = SA_Temperature(
                z,
                self.surface_temperature,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
            )
        return temperature

    def pressure(self, z):
        if self.source in ["netCDF", "GFS"]:
            pressure = interp_ex(z, self.z_data, self.pressure_data)
        elif self.source == "standardAtmos":
            pressure = SA_Pressure(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return pressure

    def density(self, z):
        if self.source in ["netCDF", "GFS"]:
            density = interp_ex(z, self.z_data, self.density_data)
        elif self.source == "standardAtmos":
            density = SA_Density(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return density

    def wind_U(self, z, scale=None):
        u = interp_ex(z, self.z_data, self.wind_U_data)
        if scale is None:
            ret = u
        else:
            ret = u / scale
        return ret

    def wind_V(self, z, scale=None):
        v = interp_ex(z, self.z_data, self.wind_V_data)
        if scale is None:
            ret = v
        else:
            ret = v / scale
        return ret

    def wind_speed(self, z):
        u = self.wind_U(z)
        v = self.wind_V(z)
        spd = np.sqrt(u * u + v * v)
        return spd

    def max_wind_speed(self, H, num=100):
        z = np.arange(0.0, H + 1.0, H / (num - 1), dtype=np.float64)
        spd = self.wind_speed(z)
        return np.amax(spd)

    def settling_speed(self, params, z, scale=None):
        Ws = self.calculate_settling_speed(params, z)
        if scale is None:
            ret = Ws
        else:
            ret = Ws / scale
        return ret

    @staticmethod
    def _solve_settling_function(diam, g, rho_p, rho_a, mu, max_iter=50):

        tolerance = 1e-12

        x0 = np.float64(1e-6)
        x1 = np.float64(10000)

        def _settling_func_white(Re, d, g, rho_p, rho_a, mu):
            gp = (rho_p - rho_a) * g / rho_a
            C1 = 0.25
            C2 = 6.0
            Cd = C1 + 24.0 / Re + C2 / (1 + np.sqrt(Re))
            d3 = d**3
            f = Cd * Re * Re - 4.0 / 3.0 * gp * d3 * (rho_a / mu) ** 2
            return f

        fx0 = _settling_func_white(x0, diam, g, rho_p, rho_a, mu)
        fx1 = _settling_func_white(x1, diam, g, rho_p, rho_a, mu)

        if np.abs(fx0) < np.abs(fx1):
            x0, x1 = x1, x0
            fx0, fx1 = fx1, fx0

        x2, fx2 = x0, fx0
        d = x2

        mflag = np.bool(True)
        steps_taken = 0

        while steps_taken < max_iter and np.abs(x1 - x0) > tolerance:
            fx0 = _settling_func_white(x0, diam, g, rho_p, rho_a, mu)
            fx1 = _settling_func_white(x1, diam, g, rho_p, rho_a, mu)
            fx2 = _settling_func_white(x2, diam, g, rho_p, rho_a, mu)

            if fx0 != fx2 and fx1 != fx2:
                L0 = (x0 * fx1 * fx2) / ((fx0 - fx1) * (fx0 - fx2))
                L1 = (x1 * fx0 * fx2) / ((fx1 - fx0) * (fx1 - fx2))
                L2 = (x2 * fx1 * fx0) / ((fx2 - fx0) * (fx2 - fx1))
                new = L0 + L1 + L2
            else:
                new = x1 - fx1 * (x1 - x0) / (fx1 - fx0)

            if (
                (new < 0.25 * (3 * x0 + x1) or new > x1)
                or (mflag and np.abs(new - x1) >= 0.5 * np.abs(x1 - x2))
                or ((not mflag) and np.abs(new - x1) >= 0.5 * np.abs(x2 - d))
                or (mflag and np.abs(x1 - x2) < tolerance)
                or ((not mflag) and np.abs(x2 - d) < tolerance)
            ):
                new = 0.5 * (x0 + x1)
                mflag = np.bool(True)
            else:
                mflag = np.bool(False)

            fnew = _settling_func_white(new, diam, g, rho_p, rho_a, mu)
            d, x2 = x2, x1

            if fx0 * fnew < 0:
                x1 = new
            else:
                x0 = new

            if np.abs(fx0) < np.abs(fx1):
                x0, x1 = x1, x0

            steps_taken += 1

        return x1, steps_taken

    def calculate_settling_speed(self, params, z):

        ws = np.empty((z.size, params.grains.bins), dtype=np.float64)

        rho_a = np.empty((z.size, 1), dtype=np.float64)
        rho_a = self.density(z)

        for iz, rho_az in enumerate(rho_a):

            for j, (d, rho_p) in enumerate(
                zip(params.grains.diameter, params.grains.density)
            ):

                Re, steps_taken = self._solve_settling_function(
                    d, params.physical.g, rho_p, rho_az, params.physical.mu
                )
                if steps_taken > 50:
                    raise RuntimeError(
                        "In MetData: _solve_settling_function" + " failed to converge"
                    )
                ws[iz, j] = params.physical.mu * Re / rho_az / d

        return ws


# pylint: disable=E1101
MetData_type = MetData.class_type.instance_type


def save_met(met, file="meteorology.npz"):
    if os.path.exists(file):
        print(
            "WARNING: {outname} ".format(outname=file)
            + "already exists and will be replaced"
        )

    if met.source == "standardAtmos":
        np.savez(
            file,
            source=met.source,
            surface_temperature=met.surface_temperature,
            surface_pressure=met.surface_pressure,
            lapse_tropos=met.lapse_tropos,
            lapse_stratos=met.lapse_stratos,
            height_tropos=met.height_tropos,
            height_stratos=met.height_stratos,
            Ra=met.Ra,
            g=met.g,
            z_data=met.z_data,
            wind_U_data=met.wind_U_data,
            wind_V_data=met.wind_V_data,
        )
    elif met.source in ["netCDF", "GFS"]:
        np.savez(
            file,
            source=met.source,
            Ra=met.Ra,
            g=met.g,
            z_data=met.z_data,
            wind_U_data=met.wind_U_data,
            wind_V_data=met.wind_V_data,
            temperature_data=met.temperature_data,
            pressure_data=met.pressure_data,
            density_data=met.density_data,
        )


def load_met(met_file):

    if not os.path.exists(met_file):
        raise IOError("AshDisperse meteorological file {} not found".format(met_file))

    met = MetData()
    data = np.load(met_file)

    if str(data["source"]) == "standardAtmos":
        met = MetData(
            source="standardAtmos",
            surface_temperature=np.float64(data["surface_temperature"]),
            surface_pressure=np.float64(data["surface_pressure"]),
            lapse_tropos=np.float64(data["lapse_tropos"]),
            lapse_stratos=np.float64(data["lapse_stratos"]),
            height_tropos=np.float64(data["height_tropos"]),
            height_stratos=np.float64(data["height_stratos"]),
            Ra=np.float64(data["Ra"]),
            g=np.float64(data["g"]),
        )
        met.set_zuv_data(data["Z"], data["U"], data["V"])
    elif str(data["source"]) in ["netCDF", "GFS"]:
        met = MetData(
            source="netCDF", Ra=np.float64(data["Ra"]), g=np.float64(data["g"])
        )
        met.z_data = data["z_data"]
        met.wind_U_data = data["wind_U_data"]
        met.wind_V_data = data["wind_V_data"]
        met.temperature_data = data["temperature_data"]
        met.pressure_data = data["pressure_data"]
        met.density_data = data["density_data"]

    return met


def netCDF_to_Met(met, netcdf_data):
    N = len(netcdf_data.altitude)
    met.z_data = np.empty((N), dtype=np.float64)
    met.z_data[:N] = netcdf_data.altitude[:N]
    met.wind_U_data = np.empty((N), dtype=np.float64)
    met.wind_U_data[:N] = netcdf_data.wind_U[:N]
    met.wind_V_data = np.empty((N), dtype=np.float64)
    met.wind_V_data[:N] = netcdf_data.wind_V[:N]
    met.temperature_data = np.empty((N), dtype=np.float64)
    met.temperature_data[:N] = netcdf_data.temperature[:N]
    met.pressure_data = np.empty((N), dtype=np.float64)
    met.pressure_data[:N] = netcdf_data.pressure[:N]
    met.density_data = np.empty((N), dtype=np.float64)
    met.density_data[:N] = netcdf_data.density[:N]
    met.source = "netCDF"
    return met


def gfs_to_Met(met, gfs_data, Ra=287.058):
    z = gfs_data.variables["Geopotential_height_isobaric"]
    temp = gfs_data.variables["Temperature_isobaric"]
    u = gfs_data.variables["u-component_of_wind_isobaric"]
    v = gfs_data.variables["v-component_of_wind_isobaric"]
    pres_name = temp.coordinates.split()[-1]
    pres = gfs_data.variables[pres_name]

    N = gfs_data.variables[pres_name].size

    met.z_data = np.empty((N), dtype=np.float64)
    met.z_data[:N] = np.flipud(z[:].data.flatten())

    met.wind_U_data = np.empty((N), dtype=np.float64)
    met.wind_U_data[:N] = np.flipud(u[:].data.flatten())

    met.wind_V_data = np.empty((N), dtype=np.float64)
    met.wind_V_data[:N] = np.flipud(v[:].data.flatten())

    met.temperature_data = np.empty((N), dtype=np.float64)
    met.temperature_data[:N] = np.flipud(temp[:].data.flatten())

    met.pressure_data = np.empty((N), dtype=np.float64)
    met.pressure_data[:N] = np.flipud(pres[:].data.flatten())

    met.density_data = np.empty((N), dtype=np.float64)
    met.density_data[:N] = met.pressure_data[:N] / Ra / met.temperature_data[:N]

    met.source = "GFS"
    return met


class netCDF:
    def __init__(self, file, lat, lon, Ra=287.058, g=9.80665):
        self.file = file
        self.lat = lat
        self.lon = lon
        self.Ra = Ra
        self.g = g

        self.altitude = np.empty((0), dtype=np.float64)
        self.wind_U = np.empty((0), dtype=np.float64)
        self.wind_V = np.empty((0), dtype=np.float64)
        self.temperature = np.empty((0), dtype=np.float64)
        self.pressure = np.empty((0), dtype=np.float64)
        self.density = np.empty((0), dtype=np.float64)

    def _near_lat_lon(self, lats, lons):
        lat_i_near = np.abs(lats - self.lat).argmin()
        lon_i_near = np.abs(lons - self.lon).argmin()

        if self.lat in lats:
            lat_i0 = lat_i_near
            lat_i1 = lat_i_near  # won't need this
        else:
            lat_i0 = lat_i_near if lat_i_near < self.lat else lat_i_near + 1
            lat_i1 = lat_i0 - 1
        if self.lon in lons:
            lon_i0 = lon_i_near
            lon_i1 = lon_i_near  # won't need this
        else:
            lon_i0 = lon_i_near if lon_i_near < self.lon else lon_i_near - 1
            lon_i1 = lon_i0 + 1

        return lat_i0, lat_i1, lon_i0, lon_i1

    def _interp_latlon(self, data, lats, lons):

        lat_i0, lat_i1, lon_i0, lon_i1 = self._near_lat_lon(lats, lons)

        d_lat = self.lat - lats[lat_i0]
        d_lon = self.lon - lons[lon_i0]

        data_00 = data[0, :, lat_i0, lon_i0]
        data_01 = data[0, :, lat_i1, lon_i0]
        data_10 = data[0, :, lat_i0, lon_i1]
        data_11 = data[0, :, lat_i1, lon_i1]

        data_i0 = data_00 + d_lon * (data_10 - data_00)
        data_i1 = data_01 + d_lon * (data_11 - data_01)
        data_ij = data_i0 + d_lat * (data_i1 - data_i0)

        return data_ij

    def extract(self):
        nc = Dataset(self.file, "r")
        P = nc.variables["level"][:] * 100

        lats = nc.variables["latitude"][:].data
        lons = nc.variables["longitude"][:].data

        if np.amax(lons) > 180.0:
            lons[lons > 180.0] = 180.0 - lons[lons > 180.0]

        lat_i0, _, lon_i0, _ = self._near_lat_lon(lats, lons)

        Z = nc.variables["z"][0, :, lat_i0, lon_i0] / self.g
        self.altitude = np.float64(np.asarray(np.flipud(Z)))

        Uij = self._interp_latlon(nc.variables["u"], lats, lons)
        Vij = self._interp_latlon(nc.variables["v"], lats, lons)
        Tij = self._interp_latlon(nc.variables["t"], lats, lons)

        self.wind_U = np.float64(np.asarray(np.flipud(Uij)))
        self.wind_V = np.float64(np.asarray(np.flipud(Vij)))
        self.temperature = np.float64(np.asarray(np.flipud(Tij)))

        self.pressure = np.float64(np.asarray(np.flipud(P)))
        self.density = np.float64(np.asarray(np.flipud(P / self.Ra / Tij)))

    def download(self, *, datetime):
        if os.path.isfile(self.file):
            print_text("Met file {} exists and will be overwritten".format(self.file))

        cds = cdsapi.Client()
        cds.retrieve(
            "reanalysis-era5-pressure-levels",
            {
                "product_type": "reanalysis",
                "format": "netcdf",
                "variable": [
                    "geopotential",
                    "temperature",
                    "u_component_of_wind",
                    "v_component_of_wind",
                ],
                "pressure_level": [
                    "1",
                    "2",
                    "3",
                    "5",
                    "7",
                    "10",
                    "20",
                    "30",
                    "50",
                    "70",
                    "100",
                    "125",
                    "150",
                    "175",
                    "200",
                    "225",
                    "250",
                    "300",
                    "350",
                    "400",
                    "450",
                    "500",
                    "550",
                    "600",
                    "650",
                    "700",
                    "750",
                    "775",
                    "800",
                    "825",
                    "850",
                    "875",
                    "900",
                    "925",
                    "950",
                    "975",
                    "1000",
                ],
                "year": str(datetime.year),
                "month": "{:02d}".format(datetime.month),
                "day": "{:02d}".format(datetime.day),
                "time": "{:02d}:00".format(datetime.hour),
                "area": [
                    ceil(self.lat + 0.25),
                    floor(self.lon - 0.25),
                    floor(self.lat - 0.25),
                    ceil(self.lon + 0.25),
                ],
            },
            self.file,
        )


def wind_scale(max_speed):

    max_log10 = np.log10(max_speed)
    major = []
    minor = []
    for j in range(np.int(max_log10) + 1):
        major.append(10**j)
        for k in range(2, 10):
            if k * 10**j > max_speed:
                break
            minor.append(k * 10**j)

    return major, minor


def wind_plot(met, z, show=True, savename=None):

    fig, ax = plt.subplots()

    U = met.wind_U(z)
    V = met.wind_V(z)
    speed = met.wind_speed(z)

    max_speed = speed.max()
    major, minor = wind_scale(max_speed)
    major.append(minor[-1])
    minor.pop(-1)

    ax.set_xlim(-max_speed, max_speed)
    ax.set_ylim(-max_speed, max_speed)
    ax.set_aspect("equal")
    ax.axis("off")

    cmap = plt.cm.viridis
    norm = Normalize(vmin=0, vmax=z[-1])
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=Normalize(vmin=0, vmax=z[-1]))

    ax.quiver(
        0,
        -max_speed,
        0,
        2 * max_speed,
        color="darkgray",
        headwidth=6,
        headlength=10,
        scale=1,
        scale_units="xy",
        angles="xy",
    )
    ax.quiver(
        -max_speed / 2,
        0,
        max_speed,
        0,
        color="darkgray",
        headwidth=1,
        headlength=0,
        scale=1,
        scale_units="xy",
        angles="xy",
    )
    ax.text(0, 1.1 * max_speed, "N")

    for m in major:
        ax.add_artist(plt.Circle((0, 0), m, ec="darkgray", lw=0.5, fill=False))
        ax.text(m * np.cos(45 * np.pi / 180), m * np.sin(45 * np.pi / 180), str(m))
    for m in minor:
        ax.add_artist(plt.Circle((0, 0), m, ec="darkgray", lw=0.25, fill=False))

    for (this_z, this_u, this_v) in zip(z, U, V):
        ax.quiver(
            0,
            0,
            this_u,
            this_v,
            color=cmap(norm(this_z)),
            scale=1,
            scale_units="xy",
            angles="xy",
        )

    fmt = ticker.FuncFormatter(lambda z, pos: "{:g}".format(z * 1e-3))
    cbar = fig.colorbar(sm, format=fmt)
    cbar.ax.set_title("Altitude (km)")

    if savename is not None:
        plt.savefig(savename)

    if show:
        plt.show()
    else:
        plt.close()

    return



from collections import OrderedDict
from numba.experimental import jitclass
from numba import complex128, float64, int64
import numpy as np

from ..spectral.cheb import ChebMat


output_spec = OrderedDict()
output_spec['start'] = float64                 # start altitude
output_spec['stop'] = float64                  # stop altitude
output_spec['step'] = float64                  # step altitude
output_spec['altitudes'] = float64[::1]        # Output altitudes
output_spec['Nz'] = int64                      # number of output altitudes
# Chebyshev matrices
output_spec['Cheb_lower'] = complex128[:, ::1]  # at lower altitudes
output_spec['Cheb_upper'] = complex128[:, ::1]  # at upper altitudes


@jitclass(output_spec)
class OutputParameters():
    def __init__(self, start=0, stop=40e3, step=1e3):

        self.start = np.float64(start)
        self.stop = np.float(stop)
        self.step = np.float(step)

        self.set_altitudes()

    def set_altitudes(self):
        if self.stop == self.start:
            self.altitudes = np.asarray([0], dtype=np.float64)
        else:
            self.altitudes = np.arange(self.start, self.stop+self.step,
                                       self.step, dtype=np.float64)
        self.Nz = len(self.altitudes)

    def ChebMats(self, N, H):
        z = self.altitudes
        z_lower = z[z <= H]
        z_upper = z[z > H]

        x_lower = 2*z_lower/H - 1
        x_upper = 1 - 2*H/z_upper

        self.Cheb_lower = ChebMat(N, x_lower)
        self.Cheb_upper = ChebMat(N, x_upper)

    def describe(self):
        print("Output parameters for AshDisperse")
        print("  Altitudes (m): ")
        for z in self.altitudes:
            print('    ', z)
        print("********************")


# pylint: disable=E1101
OutputParameters_type = OutputParameters.class_type.instance_type



from collections import OrderedDict
from numba.experimental import jitclass
from numba import float64
import numpy as np
from math import ceil


model_spec = OrderedDict()
# Scales and dimensionless parameters as arrays, one for each grain size
model_spec['SettlingScale'] = float64[::1]    # Settling speed
model_spec['Velocity_ratio'] = float64[::1]   # wind speed/settling speed
model_spec['xyScale'] = float64[::1]          # Horizontal scale
model_spec['Lx'] = float64[::1]               # Dimensionless extent in x
model_spec['Ly'] = float64[::1]               # Dimensionless extent in y
model_spec['cScale'] = float64[::1]           # Concentration
model_spec['QScale'] = float64[::1]           # Source term
model_spec['Peclet_number'] = float64         # Peclet number
model_spec['Diffusion_ratio'] = float64       # Kappa_h/Kappa_v
model_spec['sigma_hat'] = float64[::1]        # source radius/xyScale
model_spec['sigma_hat_scale'] = float64[::1]  # sigma_hat scaled to -pi->pi


@jitclass(model_spec)
class ModelParameters():

    # pylint: disable=too-many-instance-attributes
    # These attributes are all needed in ModelParameters

    def __init__(self):
        self.SettlingScale = np.empty((1), dtype=np.float64)
        self.SettlingScale[:] = np.nan
        self.Velocity_ratio = np.empty((1), dtype=np.float64)
        self.Velocity_ratio[:] = np.nan
        self.xyScale = np.empty((1), dtype=np.float64)
        self.xyScale[:] = np.nan
        self.Lx = np.empty((1), dtype=np.float64)
        self.Lx[:] = np.nan
        self.Ly = np.empty((1), dtype=np.float64)
        self.Ly[:] = np.nan
        self.cScale = np.empty((1), dtype=np.float64)
        self.cScale[:] = np.nan
        self.QScale = np.empty((1), dtype=np.float64)
        self.QScale[:] = np.nan
        self.Peclet_number = np.nan
        self.Diffusion_ratio = np.nan
        self.sigma_hat = np.empty((1), dtype=np.float64)
        self.sigma_hat[:] = np.nan
        self.sigma_hat_scale = np.empty((1), dtype=np.float64)
        self.sigma_hat_scale[:] = np.nan

    def from_params(self, solver_params, met_params, source_params,
                    grain_params, physical_params):
        N = met_params.Ws_scale.shape[0]
        self.SettlingScale = np.empty((N), dtype=np.float64)
        self.Velocity_ratio = np.empty((N), dtype=np.float64)
        self.xyScale = np.empty((N), dtype=np.float64)
        self.Lx = np.empty((N), dtype=np.float64)
        self.Ly = np.empty((N), dtype=np.float64)
        self.cScale = np.empty((N), dtype=np.float64)
        self.QScale = np.empty((N), dtype=np.float64)
        self.sigma_hat = np.empty((N), dtype=np.float64)
        self.sigma_hat_scale = np.empty((N), dtype=np.float64)

        self.Peclet_number = (physical_params.Kappa_h
                              / met_params.U_scale / source_params.PlumeHeight)
        self.Diffusion_ratio = physical_params.Kappa_h/physical_params.Kappa_v
        for j in range(N):
            self.SettlingScale[j] = met_params.Ws_scale[j]
            self.Velocity_ratio[j] = met_params.U_scale/self.SettlingScale[j]
            self.xyScale[j] = (met_params.U_scale*source_params.PlumeHeight
                               / self.SettlingScale[j])
            self.Lx[j] = (ceil(3*source_params.radius/self.xyScale[j])
                          * solver_params.domX)
            self.Ly[j] = (ceil(3*source_params.radius/self.xyScale[j])
                          * solver_params.domY)
            self.cScale[j] = (grain_params.proportion[j]*source_params.MER
                              / source_params.radius**2
                              / self.SettlingScale[j])
            self.QScale[j] = grain_params.proportion[j]*source_params.MER
            self.sigma_hat[j] = source_params.radius/self.xyScale[j]
            self.sigma_hat_scale[j] = (np.pi*self.sigma_hat[j]
                                       / np.sqrt(solver_params.domX
                                                 * solver_params.domY))

    def from_values(self, SettlingScale, Velocity_ratio,
                    xyScale, Lx, Ly, cScale,
                    QScale, Peclet_number, Diffusion_ratio, sigma_hat,
                    sigma_hat_scale):
        self.SettlingScale = SettlingScale
        self.Velocity_ratio = Velocity_ratio
        self.xyScale = xyScale
        self.Lx = Lx
        self.Ly = Ly
        self.cScale = cScale
        self.QScale = QScale
        self.Peclet_number = Peclet_number
        self.Diffusion_ratio = Diffusion_ratio
        self.sigma_hat = sigma_hat
        self.sigma_hat_scale = sigma_hat_scale

    def describe(self):
        print("Model parameters for AshDisperse")
        print("  Settling speed scale = ", self.SettlingScale)
        print("  Velocity ratio = ", self.Velocity_ratio)
        print("  concentration scale = ", self.cScale)
        print("  x and y scale = ", self.xyScale)
        print("  Lx = ", self.Lx)
        print("  Ly = ", self.Ly)
        print("  source flux scale = ", self.QScale)
        print("  Peclet number = ", self.Peclet_number)
        print("  Diffusion ratio = ", self.Diffusion_ratio)
        print("********************")


# pylint: disable=E1101
ModelParameters_type = ModelParameters.class_type.instance_type



from collections import OrderedDict
from numba.experimental import jitclass
from numba import float64, int64
import numpy as np


solver_spec = OrderedDict()
solver_spec['domX'] = float64     # dimensionless domain length in x
solver_spec['domY'] = float64     # dimensionless domain length in y
solver_spec['minN_log2'] = int64  # log2 minimum number of Chebyshev points
solver_spec['minN'] = int64       # minimum number of Chebyshev points
solver_spec['maxN_log2'] = int64  # log2 maximum number of Chebyshev points
solver_spec['maxN'] = int64       # maximum number of Chebyshev points
solver_spec['Nx_log2'] = int64    # log2 of number of points in x
solver_spec['Nx'] = int64         # number of points in x
solver_spec['Ny_log2'] = int64    # log2 of number of points in y
solver_spec['Ny'] = int64         # number of points in y
solver_spec['chebIts'] = int64    # number of chebyshev iterates
solver_spec['epsilon'] = float64  # tolerance for converged spectral series
solver_spec['meps'] = float64     # Machine epsilon


@jitclass(solver_spec)
class SolverParameters():
    def __init__(self, domX=1.5, domY=1.5, minN_log2=4, maxN_log2=8, Nx_log2=8,
                 Ny_log2=8, epsilon=1e-8):
        self.meps = np.finfo(np.float64).eps

        if domX < 0:
            raise ValueError('In SolverParameters, must have domX>0')
        self.domX = np.float64(domX)  # Dimensionless domain size in x

        if domY < 0:
            raise ValueError('In SolverParameters, must have domY>0')
        self.domY = np.float64(domY)  # Dimensionless domain size in y

        if minN_log2 < 0:
            raise ValueError('In SolverParameters, must have minN_log2>0')
        self.minN_log2 = np.int64(minN_log2)  # Minimum z-resolution (log2)
        self.minN = 2**self.minN_log2

        if maxN_log2 < 0:
            raise ValueError('In SolverParameters, must have maxN_log2>0')
        self.maxN_log2 = np.int64(maxN_log2)  # Maximum z-resolution (log2)
        self.maxN = 2**self.maxN_log2

        if minN_log2 > maxN_log2:
            raise ValueError(
                'In SolverParameters, must have minN_log2 < maxN_log2')

        if Nx_log2 < 0:
            raise ValueError('In SolverParameters, must have Nx_log2>0')
        self.Nx_log2 = np.int64(Nx_log2)  # x-resolution (log2)
        self.Nx = 2**self.Nx_log2

        if Ny_log2 < 0:
            raise ValueError('In SolverParameters, must have Ny_log2>0')
        self.Ny_log2 = np.int64(Ny_log2)  # y-resolution (log2)
        self.Ny = 2**self.Ny_log2

        self.chebIts = self.maxN_log2 - self.minN_log2 + 1

        if epsilon < 0:
            raise ValueError('In SolverParameters, must have epsilon>0')
        if epsilon < 10*self.meps:
            raise ValueError(
                'In SolverParameters, '
                + 'must have epsilon >= 10*machine epsilon')
        self.epsilon = np.float64(epsilon)

    def describe(self):
        print("Solver parameters for AshDisperse")
        print("  Dimensionless domain size in x, domX = ", self.domX)
        print("  Dimensionless domain size in y, domY = ", self.domY)
        print("  Minimum resolution in z, minN = ", self.minN,
              " (minN_log2 = ", self.minN_log2, ")")
        print("  Maximum resolution in z, maxN = ", self.maxN,
              " (maxN_log2 = ", self.maxN_log2, ")")
        print("  Number of Chebyshev iterates = ", self.chebIts)
        print("  Tolerance for Chebyshev series, epsilon = ", self.epsilon)
        print("  Resolution in x, Nx = ", self.Nx,
              " (Nx_log2 = ", self.Nx_log2, ")")
        print("  Resolution in y, Ny = ", self.Ny,
              " (Ny_log2 = ", self.Ny_log2, ")")
        print("********************")


# pylint: disable=E1101
SolverParameters_type = SolverParameters.class_type.instance_type



import os
import datetime
from collections import OrderedDict
from numba.experimental import jitclass
from numba import optional
import numpy as np

from .solver_params import (SolverParameters, SolverParameters_type)
from .grain_params import (GrainParameters, GrainParameters_type)
from .source_params import (SourceParameters, SourceParameters_type)
from .physical_params import (PhysicalParameters, PhysicalParameters_type)
from .met_params import (MetParameters, MetParameters_type)
from .model_params import (ModelParameters, ModelParameters_type)
from .output_params import (OutputParameters, OutputParameters_type)

from ..queryreport import is_valid_lat
from ..queryreport import is_valid_lon

from ..utilities import latlon_point_to_utm_code


param_spec = OrderedDict()
param_spec['solver'] = optional(SolverParameters_type)
param_spec['grains'] = optional(GrainParameters_type)
param_spec['source'] = optional(SourceParameters_type)
param_spec['physical'] = optional(PhysicalParameters_type)
param_spec['met'] = optional(MetParameters_type)
param_spec['model'] = optional(ModelParameters_type)
param_spec['output'] = optional(OutputParameters_type)


@jitclass(param_spec)
class Parameters():
    def __init__(self):
        self.solver = None
        self.grains = None
        self.source = None
        self.physical = None
        self.met = None
        self.model = None
        self.output = None

    def describe(self):
        print("AshDisperse parameters")
        self.source.describe()
        self.grains.describe()
        self.solver.describe()
        self.physical.describe()
        # self.met.describe()
        self.model.describe()
        self.output.describe()


# pylint: disable=E1101
Parameters_type = Parameters.class_type.instance_type


def copy_parameters(A):
    new_A = Parameters()
    new_A.solver = SolverParameters(
        domX=A.solver.domX,
        domY=A.solver.domY,
        minN_log2=A.solver.minN_log2,
        maxN_log2=A.solver.maxN_log2,
        epsilon=A.solver.epsilon,
        Nx_log2=A.solver.Nx_log2,
        Ny_log2=A.solver.Ny_log2)
    new_A.grains = GrainParameters()
    for j in range(A.grains.bins):
        new_A.grains.add_grain(A.grains.diameter[j],
                               A.grains.density[j],
                               A.grains.proportion[j])
    new_A.source = SourceParameters(
        A.source.latitude,
        A.source.longitude,
        A.source.utmcode,
        A.source.radius,
        A.source.PlumeHeight,
        A.source.MER,
        A.source.duration,
        A.source.Suzuki_k,
        None)
    new_A.physical = PhysicalParameters(
        A.physical.Kappa_h,
        A.physical.Kappa_v,
        A.physical.g,
        A.physical.mu)
    new_A.met = MetParameters(A.met.U_scale, A.met.Ws_scale)
    new_A.model = ModelParameters()
    new_A.model.from_values(
        A.model.SettlingScale,
        A.model.Velocity_ratio,
        A.model.xyScale,
        A.model.Lx,
        A.model.Ly,
        A.model.cScale,
        A.model.QScale,
        A.model.Peclet_number,
        A.model.Diffusion_ratio,
        A.model.sigma_hat,
        A.model.sigma_hat_scale)
    new_A.output = OutputParameters(
        A.output.start,
        A.output.stop,
        A.output.step)
    new_A.output.set_altitudes()
    new_A.output.ChebMats(A.solver.maxN, A.source.PlumeHeight)
    return new_A


def update_parameters(A, met,
                      domX=None,
                      domY=None,
                      minN_log2=None,
                      maxN_log2=None,
                      epsilon=None,
                      Nx_log2=None,
                      Ny_log2=None,
                      grains=None,
                      latitude=None,
                      longitude=None,
                      radius=None,
                      PlumeHeight=None,
                      MER=None,
                      duration=None,
                      Suzuki_k=None,
                      Suzuki_peak=None,
                      Kappa_h=None,
                      Kappa_v=None,
                      g=None,
                      mu=None,
                      U_scale=None,
                      Ws_scale=None,
                      start=None,
                      stop=None,
                      step=None):

    # solver
    if domX is not None:
        A.solver.domX = domX
    if domY is not None:
        A.solver.domY = domY
    if minN_log2 is not None:
        A.solver.minN_log2 = minN_log2
        A.solver.minN = 2**minN_log2
    if maxN_log2 is not None:
        A.solver.maxN_log2 = maxN_log2
        A.solver.maxN = 2**maxN_log2
    A.solver.chebIts = A.solver.maxN_log2-A.solver.minN_log2 + 1
    if epsilon is not None:
        A.solver.epsilon = epsilon
    if Nx_log2 is not None:
        A.solver.Nx_log2 = Nx_log2
        A.solver.Nx = 2**Nx_log2
    if Ny_log2 is not None:
        A.solver.Ny_log2 = Ny_log2
        A.solver.Ny = 2**Ny_log2

    # grains
    if grains is not None:
        if not isinstance(grains, list):
            raise ValueError(
                "in update_parameters, grains must be a list of dicts; "
                + "received {}".format(grains))
        if not all('class' in g and 'diameter' in g
                   and 'density' in g and 'proportion' in g for g in grains):
            raise ValueError(
                "in update parameters, grains must be a list of dicts with "
                + "dict containing the keys 'class', 'diameter', 'density',"
                + " 'proportion'; received {}".format(grains))
        grains = sorted(grains, key=lambda p: p['proportion'], reverse=True)
        for grain in grains:
            grain_i = grain['class'] - 1
            Diameter = grain['diameter']
            Density = grain['density']
            proportion = grain['proportion']
            if grain_i < A.grains.bins:
                if proportion > 0:
                    A.grains.diameter[grain_i] = Diameter
                    A.grains.density[grain_i] = Density
                    A.grains.proportion[grain_i] = proportion
                else:
                    A.grains.remove_grain(grain_i)
            else:
                A.grains.add_grain(Diameter, Density, proportion)
        A.grains.validate()

    # source
    if latitude is not None:
        if is_valid_lat(latitude):
            A.source.latitude = np.float64(latitude)
        else:
            raise ValueError(
                "In update_parameters, latitude must be a valid latitude in "
                + " decimal degrees")
    if longitude is not None:
        if is_valid_lon(longitude):
            A.source.longitude = np.float64(longitude)
        else:
            raise ValueError(
                "In update_parameters, latitude must be a valid longitude in "
                + " decimal degrees")
    if ((latitude is not None) or (longitude is not None)):
        A.source.utmcode = latlon_point_to_utm_code(A.source.latitude,
                                                    A.source.longitude)

    if radius is not None:
        if radius < 0:
            raise ValueError('In update_parameters, radius must be positive')
        A.source.radius = np.float64(radius)

    if PlumeHeight is not None:
        if PlumeHeight < 0:
            raise ValueError(
                "In update_parameters, PlumeHeight must be positive")
        A.source.PlumeHeight = np.float64(PlumeHeight)

    if MER is not None:
        if MER < 0:
            raise ValueError('In update_parameters, MER must be positive')
        A.source.MER = np.float64(MER)

    if duration is not None:
        if duration < 0:
            raise ValueError('In update_parameters, duration must be positive')
        A.source.duration = np.float64(duration)

    if Suzuki_k is not None:
        if Suzuki_k < 1:
            raise ValueError('In update_parameters, must have Suzuki_k>=1')
        A.source.Suzuki_k = np.float64(Suzuki_k)
        A.source.peak_from_k()

    if Suzuki_peak is not None:
        if Suzuki_peak < 0 or Suzuki_peak > 1:
            raise ValueError(
                'In update_parameters, must have 0<Suzuki_peak<=1')
        A.source.Suzuki_peak = Suzuki_peak
        A.source.k_from_peak()

    # physical
    if Kappa_h is not None:
        if Kappa_h < 0:
            raise ValueError(
                'In PhysicalParameters, Kappa_h must be positive')
        A.physical.Kappa_h = np.float64(Kappa_h)

    if Kappa_v is not None:
        if Kappa_v < 0:
            raise ValueError(
                'In PhysicalParameters, Kappa_v must be positive')
        A.physical.Kappa_v = np.float64(Kappa_v)

    if g is not None:
        if g <= 0:
            raise ValueError('In PhysicalParameters, g must be positive')
        A.physical.g = np.float64(g)

    if mu is not None:
        if mu <= 0:
            raise ValueError('In PhysicalParameters, mu must be positive')
        A.physical.mu = np.float64(mu)

    wind_speed = met.max_wind_speed(A.source.PlumeHeight)
    ws = met.settling_speed(A, np.atleast_1d(0.0))[0]
    A.met = MetParameters(wind_speed, ws)

    A.model.from_params(A.solver, A.met, A.source, A.grains, A.physical)

    if start is not None:
        A.output.start = np.float64(start)
    if stop is not None:
        A.output.stop = np.float(stop)
    if step is not None:
        A.output.step = np.float(step)

    A.output.set_altitudes()
    A.output.ChebMats(A.solver.maxN, A.source.PlumeHeight)

    return A


def save_parameters(params, file='parameters.txt'):
    if os.path.exists(file):
        print('WARNING: {outname} already exists and will be replaced'
              .format(outname=file))

    with open(file, 'w') as f:
        f.write('AshDisperse parameters {}\n\n'
                .format(datetime.datetime.now().strftime("%c")))
        # source
        f.write('Source parameters\n')
        f.write('Name = {}\n'.format(params.source.name))
        f.write('Latitude = {}\n'.format(params.source.latitude))
        f.write('Longitude = {}\n'.format(params.source.longitude))
        f.write('UTM code = {}\n'.format(params.source.utmcode))
        f.write('Radius = {}\n'.format(params.source.radius))
        f.write('Plume height = {}\n'.format(params.source.PlumeHeight))
        f.write('Mass eruption rate = {}\n'.format(params.source.MER))
        f.write('Duration = {}\n'.format(params.source.duration))
        f.write('Suzuki k = {}\n'.format(params.source.Suzuki_k))
        f.write('Suzuki peak = {}\n'.format(params.source.Suzuki_peak))
        f.write('\n')

        # grains
        f.write('Grain parameters\n')
        f.write('Grain classes = {}\n'.format(params.grains.bins))
        for j in range(params.grains.bins):
            f.write('Grain {i}:(diameter={d},density={rho},proportion={p})\n'
                    .format(i=j+1, d=params.grains.diameter[j],
                            rho=params.grains.density[j],
                            p=params.grains.proportion[j]))
        f.write('\n')

        # solver
        f.write('Solver parameters\n')
        f.write('domX = {}\n'.format(params.solver.domX))
        f.write('domY = {}\n'.format(params.solver.domY))
        f.write('minN_log2 = {}\n'.format(params.solver.minN_log2))
        f.write('maxN_log2 = {}\n'.format(params.solver.maxN_log2))
        f.write('Nx_log2 = {}\n'.format(params.solver.Nx_log2))
        f.write('Nx = {}\n'.format(params.solver.Nx))
        f.write('Ny_log2 = {}\n'.format(params.solver.Ny_log2))
        f.write('Ny = {}\n'.format(params.solver.Ny))
        f.write('epsilon = {}\n'.format(params.solver.epsilon))
        f.write('\n')

        # physical
        f.write('Physical parameters\n')
        f.write('Kappa_h = {}\n'.format(params.physical.Kappa_h))
        f.write('Kappa_v = {}\n'.format(params.physical.Kappa_v))
        f.write('g = {}\n'.format(params.physical.g))
        f.write('mu = {}\n'.format(params.physical.mu))
        f.write('\n')

        # output
        f.write('Output parameters\n')
        f.write('Lower altitude = {}\n'.format(params.output.start))
        f.write('Upper altitude = {}\n'.format(params.output.stop))
        f.write('Altitude step = {}\n'.format(params.output.step))
        f.write('\n')
    return


def load_parameters(file):
    if not os.path.exists(file):
        raise IOError('AshDisperse parameters file {} not found'.format(file))

    paramslist = open(file, 'r').readlines()
    pdict = dict()
    grainlist = []
    for param in paramslist:
        param = param.rstrip('\n')
        grainval = param.split(':', 2)
        if len(grainval) == 2:
            grainnum = grainval[0].split('Grain')[1].strip(' ')
            grainprops = grainval[1].strip('()').split(',', 3)
            graindict = dict()
            for gp in grainprops:
                (prop, val) = gp.split('=', 2)
                graindict[prop.strip(' ')] = np.float64(val)
            grainlist.append(graindict)

        keyval = param.split('=', 2)
        if len(keyval) == 2:
            key = keyval[0].strip()
            value = keyval[1].strip()
            try:
                pdict[key] = np.float64(value)
            except:
                pdict[key] = value

    params = Parameters()
    params.solver = SolverParameters(domX=pdict['domX'],
                                     domY=pdict['domY'],
                                     minN_log2=pdict['minN_log2'],
                                     maxN_log2=pdict['maxN_log2'],
                                     epsilon=pdict['epsilon'],
                                     Nx_log2=pdict['Nx_log2'],
                                     Ny_log2=pdict['Ny_log2'])

    params.grains = GrainParameters()
    for j in range(int(pdict['Grain classes'])):
        params.grains.add_grain(grainlist[j]['diameter'],
                                grainlist[j]['density'],
                                grainlist[j]['proportion'])

    if 'Name' in pdict:
        Name = pdict['Name']
    else:
        Name = None
    if 'Suzuki k' in pdict:
        Suzuki_k = pdict['Suzuki k']
        Suzuki_peak = None
    elif 'Suzuki peak' in pdict:
        Suzuki_k = None
        Suzuki_peak = pdict['Suzuki peak']
    else:
        raise RuntimeError(
            "Suzuki parameters not set in {}; ".format(file)
            + "need either 'Suzuki k' or 'Suzuki peak' specified")
    params.source = SourceParameters(pdict['Latitude'],
                                     pdict['Longitude'],
                                     pdict['UTM code'],
                                     radius=pdict['Radius'],
                                     PlumeHeight=pdict['Plume height'],
                                     MER=pdict['Mass eruption rate'],
                                     duration=pdict['Duration'],
                                     Suzuki_k=Suzuki_k,
                                     Suzuki_peak=Suzuki_peak,
                                     name=Name)

    params.physical = PhysicalParameters(Kappa_h=pdict['Kappa_h'],
                                         Kappa_v=pdict['Kappa_v'],
                                         g=pdict['g'],
                                         mu=pdict['mu'])

    params.output = OutputParameters(start=pdict['Lower altitude'],
                                     stop=pdict['Upper altitude'],
                                     step=pdict['Altitude step'])
    params.output.set_altitudes()
    params.output.ChebMats(params.solver.maxN, params.source.PlumeHeight)

    # windSpeed = met.max_wind_speed(params.source.PlumeHeight)
    # ws = met.settling_speed(params, np.atleast_1d(0.0))[0]
    # params.met = MetParameters(windSpeed, ws)

    # params.model = ModelParameters()
    # params.model.from_params(params.solver,
    #                          params.met,
    #                          params.source,
    #                          params.grains,
    #                          params.physical)

    return params



from collections import OrderedDict
from numba.experimental import jitclass
from numba import float64, int64, optional
from numba.types import unicode_type
import numpy as np

source_spec = OrderedDict()
source_spec['name'] = optional(unicode_type)    # name of source
source_spec['latitude'] = float64               # latitude of source
source_spec['longitude'] = float64              # longitude of source
source_spec['utmcode'] = int64                  # EPSG utm code of source
source_spec['radius'] = float64                 # radius of the source
source_spec['PlumeHeight'] = float64            # height of the source plume
source_spec['MER'] = float64                    # mass eruption rate
source_spec['duration'] = float64               # duration of eruption
source_spec['Suzuki_k'] = optional(float64)     # Suzuki k parameter
source_spec['Suzuki_peak'] = optional(float64)  # Suzuki_peak parameter


@jitclass(source_spec)
class SourceParameters():

    def __init__(self, lat, lon, utmcode,
                 radius=10e3,
                 PlumeHeight=10e3,
                 MER=1e6,
                 duration=18000,
                 Suzuki_k=None,
                 Suzuki_peak=0.9,
                 name=None):


        self.latitude = np.float64(lat)
        self.longitude = np.float64(lon)
        self.utmcode = utmcode

        if radius < 0:
            raise ValueError('In SourceParameters, radius must be positive')
        self.radius = np.float64(radius)

        if PlumeHeight < 0:
            raise ValueError(
                'In SourceParameters, PlumeHeight must be positive')
        self.PlumeHeight = np.float64(PlumeHeight)

        if MER < 0:
            raise ValueError('In SourceParameters, MER must be positive')
        self.MER = np.float64(MER)

        if duration < 0:
            raise ValueError('In SourceParameters, duration must be positive')
        self.duration = np.float64(duration)

        if Suzuki_k is not None:
            if Suzuki_k < 1:
                raise ValueError('In SourceParameters, must have Suzuki_k>=1')
            self.Suzuki_k = np.float64(Suzuki_k)
            self.peak_from_k()

        if Suzuki_peak is not None:
            if Suzuki_peak < 0 or Suzuki_peak > 1:
                raise ValueError(
                    'In SourceParameters, must have 0<Suzuki_peak<=1')
            self.Suzuki_peak = Suzuki_peak
            self.k_from_peak()
        
        if name is not None:
            self.name = name

    def k_from_peak(self):
        self.Suzuki_k = 1./(1.-self.Suzuki_peak)

    def peak_from_k(self):
        self.Suzuki_peak = 1. - 1./self.Suzuki_k

    def describe(self):
        print("Source parameters for AshDisperse")
        print("  Mass eruption rate MER = ", self.MER, " kg/s")
        print("  Eruption duration = ", self.duration, " s")
        print("  Plume height H = ", self.PlumeHeight, " m")
        print("  Gaussian source radius = ", self.radius, " m")
        print("  Suzuki emission profile k-parameter = ", self.Suzuki_k)
        print("  Suzuki emission profile peak-parameter = ", self.Suzuki_peak)
        print("********************")


# pylint: disable=E1101
SourceParameters_type = SourceParameters.class_type.instance_type



from collections import OrderedDict
from numba.experimental import jitclass
from numba import float64
import numpy as np


phys_spec = OrderedDict()
phys_spec['Kappa_h'] = float64  # Horizontal diffusion coefficient
phys_spec['Kappa_v'] = float64  # Vertical diffusion coefficient
phys_spec['g'] = float64        # Gravitational acceleration
phys_spec['mu'] = float64       # Viscosity of air


@jitclass(phys_spec)
class PhysicalParameters():

    def __init__(self, Kappa_h=100, Kappa_v=10, g=9.81, mu=18.5e-6):

        if Kappa_h < 0:
            raise ValueError(
                'In PhysicalParameters, Kappa_h must be positive')
        self.Kappa_h = np.float64(Kappa_h)

        if Kappa_v < 0:
            raise ValueError(
                'In PhysicalParameters, Kappa_v must be positive')
        self.Kappa_v = np.float64(Kappa_v)

        if g <= 0:
            raise ValueError('In PhysicalParameters, g must be positive')
        self.g = np.float64(g)

        if mu <= 0:
            raise ValueError('In PhysicalParameters, mu must be positive')
        self.mu = np.float64(mu)

    def describe(self):
        print("Physical parameters for AshDisperse")
        print("  Horizontal diffusion coefficient Kappa_h = ",
              self.Kappa_h, " m^2/s")
        print("  Vertical diffusion coefficient Kappa_v = ",
              self.Kappa_v, " m^2/s")
        print("  Gravitational acceleration g = ",
              self.g, " m/s^2")
        print("  Viscosity of air mu = ",
              self.mu, " kg/m/s")
        print("********************")


# pylint: disable=E1101
PhysicalParameters_type = PhysicalParameters.class_type.instance_type
